import numpy as np
import matplotlib.pyplot as plt

import matplotlib as mpl
mpl.rcParams['axes.labelsize'] = "22"
mpl.rcParams['xtick.labelsize'] = "22"
mpl.rcParams['ytick.labelsize'] = "22"



colors = [(0,0,0),(0.9,0.6,0),(0.35,0.7,0.9),(0,0.6,0.5),(0.95,0.9,0.25),(0,0.45,0.7),(0.8,0.4,0),(0.8,0.6,0.7)]
symbols = ["o","s","D", "v","^"]



skeys = ["shear","buckle","stretch","propeller","stagger","opening"]
dkeys = ["shift","tilt","slide","roll","rise","twist"]
keys = skeys+dkeys
okeys = keys
degkeys = ["buckle","propeller","opening","tilt","roll","twist"]
wantedkeys = []
wskeys = []
wdkeys = []
for k in keys:
  wantedkeys.append("<"+k+">")
for k in skeys:
  wskeys.append("<"+k+">")
for k in dkeys:
  wdkeys.append("<"+k+">")

def runningMean(x, N):
  return np.convolve(x, np.ones((N,))/N)[(N-1):]

# INPUT x = values, N number blocks
# OUTPUT ss = standard deviation of standard deviations
def errBar(x,N):
  l = len(x)
  n =int(np.floor(l/N))
#  print("length: ",n)
  s = np.std(x[:N*n].reshape(N,-1),axis=1)
  ss = np.std(s)/np.sqrt(N)
  return ss

# INPUT x = (very large number of) possibly correlated x values
# OUTPUT ws = window size, err = standard error as function of window size
def checkErr(x):
  l = len(x)
  maxn = int(np.floor(l/30))
  ws = []
  err = []
  maxk = int(np.floor(np.log(maxn)/np.log(2)))
  for k in range(maxk+1):
    n = 2**k
    err.append(getErr(x,n))
    ws.append(n)
  return np.array(ws),np.array(err)

# INPUT x = array of correlated values, n = block length
# OUTPUT sErr = standard error with block length N
def getErr(x,n):
  l = len(x)
  N = int(np.floor(l/n))
  m = np.mean(x[:N*n].reshape(N,-1),axis=1)

  sErr = np.std(m)/np.sqrt(N)
  if N<20 and N>1:
    sErr= np.std(m)/np.sqrt(N-1)
  elif N==0:
    sErr=0
  return sErr

def saveShortNPZ(npyname,files,bps,name,dt,skeys=wskeys,dkeys=wdkeys):
  key = []
  dat = []
  for I in range(len(files)):
    k, d = readx3d(files[I])
    key.append(k)
    dat.append(d)
  keeps = []
  keepd = []
  for I in range(len(dat[0])):
    if key[0][I] in skeys:
      keeps.append(I)
    if key[0][I] in dkeys:
      keepd.append(I)
#  print(keep)
  
  dout = []
  kout = []
  for I in range(len(dat)):
    t = []
    k = []
    for J in range(len(keeps)):
      r = np.array(dat[I][keeps[J]])
      t.append(r)
      k.append( np.array(key[0][ keeps[J] ]) )
    for J in range(len(keepd)):
      r = dat[I][keepd[J]]
      z = np.zeros((len(r),1),dtype=float)
      s = np.concatenate((r,z),axis=1)
      t.append(s)
      k.append( np.array(key[0][ keepd[J] ]) )
    dout.append(np.array(t))
    kout.append(np.array(k))
  shiftT = 0
  for I in range(len(dout)):
    dout[I][:,:,0] = dout[I][:,:,0]+shiftT
    shiftT = dout[I][0,-1,0]
  dout2 = np.concatenate(dout,axis=1)
  kout = kout[0]
  for I in range(len(kout)):
    kout[I] = kout[I][1:-1]
  kout = np.array(kout)
  f = open(npyname, 'wb')
  np.savez_compressed(f,dout2,kout,np.array(bps),np.array(name),np.array(float(dt)))
  f.close()

# Returns the contents of expt broken out of the list -- Just a shortcut to rename the parts of expt
def unpack(expt):
  return expt[0],expt[1],expt[2],expt[3],expt[4]

# Takes a list of expts, rearranges the list, returning ordered lists of the contents of expts
def unpackList(expts):
  a = []
  for I in range(len(expts[0])):
    a.append([])
  for I in range(len(expts)):
    for J in range(len(expts[0])):
      a[J].append(expts[I][J])
  return a[0],a[1],a[2],a[3],a[4]

# Plots the values over time of an experiment. Creates a separate plot for each x3dna parameter 
# with names prefix.stretch.pdf, prefix.roll.pdf, prefix.rise.pdf, etc.
def PlotVals(expt,prefix):
  data,keys,bps,name,dt = unpack(expt)
  for I in range(len(data)):
    f, ax = plt.subplots()
    ax.set_xlabel("Time (ns)")
    plt.title("Parameter: {}".format(keys[I]))
    m,ma = get3stdev(data[I])
    print(keys[I],m,ma)
    for J in range(len(ma)):
      bp = ma[J]
      if keys[I] in skeys:
        print("bp {},stddev {:.3f},bases {}".format(bp+1,m[J],bps[bp]))
        ax.plot(data[I][:,0]*dt,data[I][:,bp],color=colors[J],label="bp {},stddev {:.3f},bases {}".format(bp+1,m[J],bps[bp]))
      elif I<len(data):
        print("step {},stddev {:.3f},bases {},{}".format(bp+1,m[J],bps[bp],bps[bp+1]))
        ax.plot(data[I][:,0]*dt,data[I][:,bp],color=colors[J],label="step {},stddev {:.3f},bases {},{}".format(bp+1,m[J],bps[bp],bps[bp+1]))
      else:
        continue
      if keys[I] in degkeys:
        ax.set_ylabel("{} parameter, (degrees)".format(keys[I]))
      else:
        ax.set_ylabel("{} parameter, (Angstroms)".format(keys[I]))
    plt.legend(prop={'size':10})
    plt.tight_layout()
    f.savefig(prefix+"."+keys[I]+".pdf")

# Takes a DATA structure, and returns the standard deviation 
# of the three data series (bases or steps) with largest standard deviations, 
# as well as which element in the structure they are.
# enddrops is the number of bases at the end of the DNA to ignore (They fluctuate a large amount and fray)
# numbases is the number (default 3) of basepairs, or steps, to return
def get3stdev(dat,enddrops=4,numbases=3):
  stdev = np.std(dat,axis=0)
  maximums = []
  maxargs = []
  for I in range(numbases):
    am = np.argmax(stdev[enddrops:-enddrops]) +enddrops
    maximums.append(stdev[am])
    stdev[am] = 0
    maxargs.append(am)
  return maximums, maxargs

# INPUT : exptref= experiment object, outpref=prefix text for the output files
# OUTPUT : writes 12 pdf files with names <outpref>.parameter.pdf
# Each pdf file has 3 parameter vs time plots for the basepairs with the highest standard deviation
# Helps to quickly check if the DNA breaks during the simulation
def plotStandard(exptref,outpref,expt=[],enddrops=4):

  dataref,keyref,bpsref,nameref,dtref = unpack(exptref)
  if len(expt)>0:
    data,keys,bps,names,dts = unpackList(expt)
  size = 30
  for I in range(len(keyref)):
    f,ax = plt.subplots()
    x,y,err = standardBases(trimData(dataref[I],dtref,endtime=80))
    x = x[enddrops:-enddrops]
    y = y[enddrops:-enddrops]
    err = err[enddrops:-enddrops]
    x2 = []
    for J in range(len(expt)):
      a,b,c = standardBases(trimData(data[J][I],dts[J]))
      a = a[enddrops:-enddrops]
      b = b[enddrops:-enddrops]
      c = c[enddrops:-enddrops]
      x2.append( (a,b,c) )
    ax.plot(x,y,color=colors[0],lw=2)
    ax.errorbar(x,y,yerr=err,color=colors[0],label="{}".format(nameref))
    
    for J in range(len(x2)):
      ax.plot(x2[J][0],x2[J][1],color=colors[J+1],lw=2)
      ax.errorbar(x2[J][0],x2[J][1],yerr=x2[J][2],color=colors[J+1],label=names[J])

    if keyref[I] in skeys:
      ax.set_xlabel("Base pair")
    else:
      ax.set_xlabel("Step")
    if keyref[I] in degkeys:
      ax.set_ylabel("Standard deviation of {} (degree)".format(keyref[I]))
    else:
      ax.set_ylabel("Standard deviation of {} (Angstrom)".format(keyref[I]))
    ax.set_xlim([enddrops,len(dataref[I][0])-enddrops])
    plt.legend(prop={'size':10})
    plt.tight_layout()
    fname = outpref+"."+keyref[I]+".pdf"
    print(fname)
    f.savefig(fname)

# INPUT : DATA with a single key
# OUTPUT: x= numbered array of base, stdev=standard deviation of all bases, 
# ss = numpy array of standard error for each basepair using 30 blocks
def standardBases(dat):
  stdev = np.std(dat[:,1:],axis=0)
  x = np.arange(len(stdev))+1
  ss = []
  for I in range(1,len(dat[0])):
    ss.append(errBar(dat[:,I],30))
  return x,stdev,np.array(ss)

# INPUT : DATA with a single key, dt=time between successive data measurements
#       : begintime = initial time in nanoseconds desired, endtime = last time desired in the trimmed output in nanoseconds
# OUTPUT: DATA with a single key that has been trimmed to start at begintime, and ending at endtime
def trimData(dat,dt,begintime=0,endtime=10000):
  return dat[np.logical_and(dat[:,0]*dt>begintime,dat[:,0]*dt<endtime)]

# Computes the percent difference in standard deviation between two trajectories 
# INPUT : dat1, dat2 = arrays of standard deviation (array iterates over basepairs)
#       : err1, err2 = arrays of standard error of standard deviation (array iterates over basepairs)
# OUTPUT: c = percent difference in standard deviation of dat1 and dat2
#       : Error of percent difference
def computePercent(dat1,dat2,err1,err2):
  tol = 0.000001
  c = (dat1/dat2-1.)*100.
  denom = dat2-err2
  denom[denom<tol] = tol
  emax = ( (dat1+err1)/(denom) -1. )*100. -c
  denom = dat2+err2
  denom[denom<tol] = tol
  emin = c - ( (dat1-err1)/(denom) -1. )*100. 
 
  return c, (emax+emin)/2.

# Computes the percent difference in standard deviation between two trajectories
# INPUT : exptref = Reference trajectory object, expt = Compared trajectory object
#       : outpref = prefix for output files, 
#       : steps = steps to analyze, bases = basepairs to analyze
# OUTPUTS files with names <outpref>.intra.pdf, <outpref>.inter.pdf for intra- and inter basepair parameters
def compPercent(exptref,expt,outpref,steps,bases,enddrops=4,timedrop=10):
  dataref,keyref,bpsref,nameref,dtref = unpack(exptref)
  data,keys,bps,names,dts = unpackList(expt)
  x = []
  ref = []
  ss = []  
  pctChange = []
  errors = []

  for I in range(len(keyref)):
    x.append([])
    ref.append([])
    ss.append([])
    x2,ref2,ss2 = standardBases(trimData(dataref[I],dtref,endtime=80))
#    print(len(dataref[I]),len(dataref[I][dataref[I][:,0]*dtref>timedrop]))
    ss2[ss2==0] = 10.
    ref2[ref2==0] = 10.
    ref[I].append(ref2)
    ss[I].append(ss2)
    x[I].append(x2)
    chg = []
    e =[]
    for J in range(len(data)):
      x2,ref2,ss2 = standardBases(trimData(data[J][I],dts[J]))
      ref[I].append(ref2)
      ss[I].append(ss2)
      x[I].append(x2)
      ref2[ref2==0] = 10.
      c,err = computePercent(ref2,ref[I][0],ss2,ss[I][0])
      chg.append(c)
      e.append(err)
    pctChange.append(chg)
    errors.append(e)
  pctChange = np.array(pctChange)
  errors = np.array(errors)    
  print(keyref)
  fs, fd = plotBars(keyref,pctChange,errors,names,steps,bases)
  fs.savefig(outpref+".intra.pdf")
  fd.savefig(outpref+".inter.pdf")

# Returns the positions in the array k where keys wk are found
# INPUT : k = array of parameters, wk = array of wanted parameters
# OUTPUT: elements = array of indices to array k, which have values included in wk
def getSubset(k, wk):
  elements = []
  for I in range(len(k)):
    if k[I] in wk:
      elements.append(I)
  return np.array(elements)

#Plots a graph of the percent change between two trajectories; Called by computePercent
# INPUT : keyref = names of parameters, pctChange = percent change (array same size as keyref)
#       : errors = error bar sizes, names = trajectory names
#       : steps, bases = steps or basepairs of interest.
def plotBars(keyref,pctChange,errors,names,steps,bases):
  print(pctChange.shape)
  els = getSubset(keyref,skeys)
  eld = getSubset(keyref,dkeys)
#  print(pctChange[els][:,0,bases])
#  for I in range(len(pctChange)):
  ys1 = np.mean(pctChange[els][:,0,bases],axis=1)
  print(ys1)
  print(skeys)
  es1 = np.sqrt(np.mean(errors[els][:,0,bases]**2,axis=1))
  yd1 = np.mean(pctChange[eld][:,0,steps],axis=1)
  ed1 = np.sqrt(np.mean(errors[eld][:,0,steps]**2,axis=1))
  
  if pctChange.shape[1]==1:
    fs, axs = plotSingle(ys1,es1,names[0],keyref[els],'')
    fd, axd = plotSingle(yd1,ed1,names[0],keyref[eld],'')
  elif pctChange.shape[1]==2:
    ys2 = np.mean(pctChange[els][:,1,bases],axis=1)
    es2 = np.sqrt(np.sum(errors[els][:,1,bases]**2,axis=1))
    yd2 = np.mean(pctChange[eld][:,1,steps],axis=1)
    ed2 = np.sqrt(np.sum(errors[eld][:,1,steps]**2,axis=1))

    fs, axs = plot2(ys1,es1,names[0],ys2,es2,names[1],keyref[els])
    fd, axd = plot2(yd1,ed1,names[0],yd2,ed2,names[1],keyref[eld])
#  plt.show()
  return fs, fd

# Creates a barchart of two data sets, returns matplotlib figure and axis objects
# INPUT : s1,s2 = data array plotted
#       : e1,e2 = size of error bars plotted
#       : lab1, lab2 = names of s1, s2... i.e. "mutant1" "mutant2"
#       : xlab = array same size as s1 that has names of each element of s1,s2. i.e. "roll","twist", etc
def plot2(s1,e1,lab1,s2,e2,lab2,xlab):
  width = 0.35
  ind = np.arange(len(s1))+0.12
  fig, ax = plt.subplots(figsize=(10,8))
  ax.bar(ind,s1,width,color=colors[2], yerr=e1,ecolor=colors[0],alpha=0.7,label=lab1)
  ax.bar(ind+width,s2,width,color=colors[3], yerr=e2,ecolor=colors[0],alpha=0.7,label=lab2)
#  ax.set_ylim([-11,22])

  ax.axhline(color='k')
  ax.set_ylabel('Delta $\sigma$ (%)')
  ax.set_xticks(ind+width)
  ax.set_xticklabels(xlab)
  ax.legend()
  plt.tight_layout()
  return fig,ax

# Creates a barchart of one data set, returns matplotlib figure and axis objects
# INPUT : s1 = data array plotted
#       : e1 = size of error bars plotted
#       : lab1 = names of s1... i.e. "mutant1" 
#       : xlab = array same size as s1 that has names of each element of s1. i.e. "roll","twist", etc
def plotSingle(s1,e1,lab1,xlab,ylab):
  width = 0.65
  ind = np.arange(len(s1))+0.12
  fig, ax = plt.subplots(figsize=(10,8))
  ax.bar(ind,s1,width,color=colors[2], yerr=e1,ecolor=colors[0],alpha=0.7,label=lab1)
  #ax.set_ylim([-11,22])

  ax.axhline(color='k')
  ax.set_ylabel('Delta $\sigma$ (%)')
  ax.set_xticks(ind+width/2)
  ax.set_xticklabels(xlab)
  ax.legend()
  plt.tight_layout()
  return fig,ax

# Reads in an npz file representing one trajectory and returns the data contained.
# INPUT : npzname = name (and path) of npz file representing a trajectory
# OUTPUT: data = time series data for a single trajectory 
#       : shape(data) = (12,T,D) where T = number of timesteps, D = number of basepairs
#       : keys = array of the names of the parameters calculated
#       : bps = names of basepairs (1 letter code A,C,T,G, etc)
#       : name = descriptive name of this trajectory.. e.g. "ATRAC"
#       : dt = timestep size
def readShortNPZ(npzname):
  npzfile = np.load(npzname)
  data = npzfile['arr_0']
  keys = npzfile['arr_1']
  bps = npzfile['arr_2']
  name = npzfile['arr_3']
  dt = npzfile['arr_4']
  return [data,keys,bps,name,dt]

# Calculates value and Propagates error for delta sigma (percent change) calculation.
# INPUT : A, B input values (B is reference, A is modification), sA, sB input errors
# OUTPUT: fraction change, error in fraction change
def propSig(A,B,sA,sB):
  sAB2 = sA*sA+sB*sB
  AmB = A-B
  SAmB = np.abs(A-B)/B*np.sqrt( sAB2/(AmB*AmB) + sB*sB/(B*B))
  return (A-B)/B,SAmB

# Reads a x3dna basepair file and returns a list of the basepairs
# INPUT : filename = name (and path) of basepair file
# OUTPUT: name = list of basepairs in the structure described by the basepair file
def readbps(filename):
  name = []
  numbp = 0
  file = open(filename,'r')
  ln = 0
  while 1:
    line = file.readline()
    ln = ln+1
    if not line:
      break
    spl = line.split()
    if ln==4:
      try:
        numbp = float(spl[0])
      except:
        print("Not a valid bps file")
        exit
    try:
      float(spl[0])
      float(spl[1])
      float(spl[2])
      float(spl[4])
    except:
      continue
    name.append(line[37]+line[49])
  return name

# Function that reads a file produced by x3dna and returns the data contained for 12 x3dna parameters
def readx3d(filename):
  exc = 0
  n = 0
  file = open(filename,'r')
  data = []
  key = []
  while 1:
    line=file.readline()
    if not line:
      break
#    print(line)
    if line[0] != '<':
      continue
    else:
      a = []
      header=line.split()[0]
      while 1:
        line=file.readline()
        if not line:
          break
        if line[:2]=='</':
          break
        line=line.split()
        lis = []
        for num in line:
          try:
            float(num)
          except:
            num=exc
          lis.append(float(num))
          n = n+1
        a.append(lis)
      if not line:
        break
      data.append(np.array(a))
      key.append(header)
#  print(len(data[0][0]))
  return key,data

